@@ -67,9 +67,15 @@
 #else
 #	include <sys/time.h>
 #	include <sys/types.h>
-#	ifdef HAVE_SYS_SELECT_H
-#		include <sys/select.h>
-#	endif
+#   if USE_POLL
+#       ifdef HAVE_SYS_POLL_H
+#	        include <sys/poll.h>
+#       endif
+#   else 
+#       ifdef HAVE_SELECT_H
+#	        include <sys/select.h>
+#       endif
+#   endif
 #endif
 
 #if LIBETPAN_IOS_DISABLE_SSL
@@ -87,12 +93,16 @@
 #  include <gnutls/x509.h>
 # endif
 # ifdef LIBETPAN_REENTRANT
-#	if HAVE_PTHREAD_H
+#	 if HAVE_PTHREAD_H
 #	  include <pthread.h>
+#  elif defined(WIN32)
+    void mailprivacy_gnupg_init_lock();
+    void mailprivacy_smime_init_lock();
+#  endif
 #	endif
-# endif
 #endif
 
+#include "mmapstring.h"
 #include "mailstream_cancel.h"
 
 struct mailstream_ssl_context
@@ -103,6 +113,9 @@ struct mailstream_ssl_context
   SSL_CTX * openssl_ssl_ctx;
   X509* client_x509;
   EVP_PKEY *client_pkey;
+# if (OPENSSL_VERSION_NUMBER >= 0x10000000L)
+  char * server_name;
+# endif /* (OPENSSL_VERSION_NUMBER >= 0x10000000L) */
 #else
   gnutls_session session;
   gnutls_x509_crt client_x509;
@@ -148,11 +161,11 @@ struct mailstream_ssl_data {
 #	define MUTEX_LOCK(x)
 #	define MUTEX_UNLOCK(x)
 #endif
-static int gnutls_init_not_required = 0;
 static int openssl_init_done = 0;
 #endif
 
 // Used to make OpenSSL thread safe
+#ifndef USE_GNUTLS
 #if defined (HAVE_PTHREAD_H) && !defined (WIN32) && defined (USE_SSL) && defined (LIBETPAN_REENTRANT)
   struct CRYPTO_dynlock_value
   {
@@ -223,26 +236,36 @@ static int openssl_init_done = 0;
     CRYPTO_set_dynlock_destroy_callback(dyn_destroy_function);
   }
 #endif
+#endif
 
 void mailstream_ssl_init_lock(void)
 {
 #if !defined (HAVE_PTHREAD_H) && defined (WIN32) && defined (USE_SSL)
-  InitializeCriticalSection(&ssl_lock);
+  static long volatile mailstream_ssl_init_lock_done = 0;
+  if (InterlockedExchange(&mailstream_ssl_init_lock_done, 1) == 0) {
+    InitializeCriticalSection(&ssl_lock);
+  }
+#endif
+}
+
+void mailstream_ssl_uninit_lock(void)
+{
+#if !defined (HAVE_PTHREAD_H) && defined (WIN32) && defined (USE_SSL)
+	static long volatile mailstream_ssl_init_lock_done = 0;
+	if (InterlockedExchange(&mailstream_ssl_init_lock_done, 1) == 0) {
+		DeleteCriticalSection(&ssl_lock);
+	}
 #endif
 }
 
 void mailstream_gnutls_init_not_required(void)
 {
-#ifdef USE_SSL
-  MUTEX_LOCK(&ssl_lock);
-  gnutls_init_not_required = 1;
-  MUTEX_UNLOCK(&ssl_lock);
-#endif
 }
 
 void mailstream_openssl_init_not_required(void)
 {
 #ifdef USE_SSL
+  mailstream_ssl_init_lock();
   MUTEX_LOCK(&ssl_lock);
   openssl_init_done = 1;
   MUTEX_UNLOCK(&ssl_lock);
@@ -258,6 +281,7 @@ void mailstream_ssl_init_not_required(void)
 static inline void mailstream_ssl_init(void)
 {
 #ifdef USE_SSL
+  mailstream_ssl_init_lock();
   MUTEX_LOCK(&ssl_lock);
 #ifndef USE_GNUTLS
   if (!openssl_init_done) {
@@ -267,15 +291,12 @@ static inline void mailstream_ssl_init(void)
     
     SSL_load_error_strings();
     SSL_library_init();
-    OpenSSL_add_all_digests();
     OpenSSL_add_all_algorithms();
-    OpenSSL_add_all_ciphers();
     
     openssl_init_done = 1;
   }
 #else
-  if (!gnutls_init_not_required)
-    gnutls_global_init();
+  gnutls_global_init();
 #endif
   MUTEX_UNLOCK(&ssl_lock);
 #endif
@@ -301,19 +322,24 @@ static inline int mailstream_prepare_fd(int fd)
 
 static int wait_SSL_connect(int s, int want_read, time_t timeout_seconds)
 {
-  fd_set fds;
   struct timeval timeout;
   int r;
-  
-  FD_ZERO(&fds);
-  FD_SET(s, &fds);
+#if defined(WIN32) || !USE_POLL
+  fd_set fds;
+#else
+  struct pollfd pfd;
+#endif // WIN32
+
   if (timeout_seconds == 0) {
     timeout = mailstream_network_delay;
   }
   else {
-		timeout.tv_sec = timeout_seconds;
+    timeout.tv_sec = timeout_seconds;
     timeout.tv_usec = 0;
   }
+#if defined(WIN32) || !USE_POLL
+  FD_ZERO(&fds);
+  FD_SET(s, &fds);
   /* TODO: how to cancel this ? */
   if (want_read)
     r = select(s + 1, &fds, NULL, NULL, &timeout);
@@ -322,11 +348,27 @@ static int wait_SSL_connect(int s, int want_read, time_t timeout_seconds)
   if (r <= 0) {
     return -1;
   }
-  
   if (!FD_ISSET(s, &fds)) {
     /* though, it's strange */
     return -1;
   }
+#else
+  pfd.fd = s;
+  if (want_read) {
+    pfd.events = POLLIN;
+  }
+  else {
+    pfd.events = POLLOUT;
+  }
+  r = poll(&pfd, 1, timeout.tv_sec * 1000 + timeout.tv_usec / 1000);
+  if (r <= 0) {
+    return -1;
+  }
+
+  if (pfd.revents & pfd.events != pfd.events) {
+    return -1;
+  }
+#endif
   
   return 0;
 }
@@ -341,6 +383,7 @@ static void mailstream_low_ssl_free(mailstream_low * s);
 static int mailstream_low_ssl_get_fd(mailstream_low * s);
 static void mailstream_low_ssl_cancel(mailstream_low * s);
 static struct mailstream_cancel * mailstream_low_ssl_get_cancel(mailstream_low * s);
+static carray * mailstream_low_ssl_get_certificate_chain(mailstream_low * s);
 
 static mailstream_low_driver local_mailstream_ssl_driver = {
   /* mailstream_read */ mailstream_low_ssl_read,
@@ -350,6 +393,10 @@ static mailstream_low_driver local_mailstream_ssl_driver = {
   /* mailstream_free */ mailstream_low_ssl_free,
   /* mailstream_cancel */ mailstream_low_ssl_cancel,
   /* mailstream_get_cancel */ mailstream_low_ssl_get_cancel,
+  /* mailstream_get_certificate_chain */ mailstream_low_ssl_get_certificate_chain,
+  /* mailstream_setup_idle */ NULL,
+  /* mailstream_unsetup_idle */ NULL,
+  /* mailstream_interrupt_idle */ NULL,
 };
 
 mailstream_low_driver * mailstream_ssl_driver = &local_mailstream_ssl_driver;
@@ -365,7 +412,7 @@ static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx)
 
 static int mailstream_openssl_client_cert_cb(SSL *ssl, X509 **x509, EVP_PKEY **pkey)
 {
-	struct mailstream_ssl_context * ssl_context = (struct mailstream_ssl_context *)SSL_CTX_get_app_data(ssl->ctx);
+	struct mailstream_ssl_context * ssl_context = (struct mailstream_ssl_context *)SSL_CTX_get_app_data(SSL_get_SSL_CTX(ssl));
 	
 	if (x509 == NULL || pkey == NULL) {
 		return 0;
@@ -384,7 +431,7 @@ static int mailstream_openssl_client_cert_cb(SSL *ssl, X509 **x509, EVP_PKEY **p
 }
 
 static struct mailstream_ssl_data * ssl_data_new_full(int fd, time_t timeout,
-	SSL_METHOD * method, void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data),
+	const SSL_METHOD * method, void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data),
 	void * cb_data)
 {
   struct mailstream_ssl_data * ssl_data;
@@ -393,6 +440,9 @@ static struct mailstream_ssl_data * ssl_data_new_full(int fd, time_t timeout,
   SSL_CTX * tmp_ctx;
   struct mailstream_cancel * cancel;
   struct mailstream_ssl_context * ssl_context = NULL;
+#if SSL_MODE_RELEASE_BUFFERS
+  long mode = 0;
+#endif
   
   mailstream_ssl_init();
   
@@ -408,9 +458,23 @@ static struct mailstream_ssl_data * ssl_data_new_full(int fd, time_t timeout,
   SSL_CTX_set_app_data(tmp_ctx, ssl_context);
   SSL_CTX_set_client_cert_cb(tmp_ctx, mailstream_openssl_client_cert_cb);
   ssl_conn = (SSL *) SSL_new(tmp_ctx);
+  
+#if SSL_MODE_RELEASE_BUFFERS
+  mode = SSL_get_mode(ssl_conn);
+  SSL_set_mode(ssl_conn, mode | SSL_MODE_RELEASE_BUFFERS);
+#endif
+  
   if (ssl_conn == NULL)
     goto free_ctx;
-  
+
+#if (OPENSSL_VERSION_NUMBER >= 0x10000000L)
+  if (ssl_context != NULL && ssl_context->server_name != NULL) {
+    SSL_set_tlsext_host_name(ssl_conn, ssl_context->server_name);
+    free(ssl_context->server_name);
+    ssl_context->server_name = NULL;
+  }
+#endif /* (OPENSSL_VERSION_NUMBER >= 0x10000000L) */
+
   if (SSL_set_fd(ssl_conn, fd) == 0)
     goto free_ssl_conn;
   
@@ -470,13 +534,19 @@ again:
 static struct mailstream_ssl_data * ssl_data_new(int fd, time_t timeout,
 	void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data), void * cb_data)
 {
-  return ssl_data_new_full(fd, timeout, SSLv23_client_method(), callback, cb_data);
-}
-
-static struct mailstream_ssl_data * tls_data_new(int fd, time_t timeout,
-  void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data), void * cb_data)
-{
-  return ssl_data_new_full(fd, timeout, TLSv1_client_method(), callback, cb_data);
+  return ssl_data_new_full(fd, timeout,
+#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
+		TLS_client_method(),
+#else
+	/* Despite their name the SSLv23_*method() functions have nothing to do
+	 * with the availability of SSLv2 or SSLv3. What these functions do is
+	 * negotiate with the peer the highest available SSL/TLS protocol version
+	 * available. The name is as it is for historic reasons. This is a very
+	 * common confusion and is the main reason why these names have been
+	 * deprecated in the latest dev version of OpenSSL. */
+		SSLv23_client_method(),
+#endif
+		callback, cb_data);
 }
 
 #else
@@ -484,10 +554,17 @@ static struct mailstream_ssl_data * tls_data_new(int fd, time_t timeout,
 static struct mailstream_ssl_context * mailstream_ssl_context_new(gnutls_session session, int fd);
 static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx);
 
+#if GNUTLS_VERSION_NUMBER <= 0x020c00
 static int mailstream_gnutls_client_cert_cb(gnutls_session session,
                                const gnutls_datum *req_ca_rdn, int nreqs,
                                const gnutls_pk_algorithm *sign_algos,
                                int sign_algos_length, gnutls_retr_st *st)
+#else
+static int mailstream_gnutls_client_cert_cb(gnutls_session session,
+                               const gnutls_datum *req_ca_rdn, int nreqs,
+                               const gnutls_pk_algorithm *sign_algos,
+                               int sign_algos_length, gnutls_retr2_st *st)
+#endif
 {
 	struct mailstream_ssl_context * ssl_context = (struct mailstream_ssl_context *)gnutls_session_get_ptr(session);
 	gnutls_certificate_type type = gnutls_certificate_type_get(session);
@@ -499,7 +576,12 @@ static int mailstream_gnutls_client_cert_cb(gnutls_session session,
 
 	if (type == GNUTLS_CRT_X509 && ssl_context->client_x509 && ssl_context->client_pkey) {
 		st->ncerts = 1;
+#if GNUTLS_VERSION_NUMBER <= 0x020c00
 		st->type = type;
+#else
+		st->cert_type = type;
+		st->key_type = GNUTLS_PRIVKEY_X509;
+#endif
 		st->cert.x509 = &(ssl_context->client_x509);
 		st->key.x509 = ssl_context->client_pkey;
 		st->deinit_all = 0;
@@ -516,6 +598,7 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, time_t timeout,
   gnutls_certificate_credentials_t xcred;
   int r;
   struct mailstream_ssl_context * ssl_context = NULL;
+  unsigned int timeout_value;
   
   mailstream_ssl_init();
   
@@ -533,8 +616,11 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, time_t timeout,
   
   gnutls_session_set_ptr(session, ssl_context);
   gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);
+#if GNUTLS_VERSION_NUMBER <= 0x020c00
   gnutls_certificate_client_set_retrieve_function(xcred, mailstream_gnutls_client_cert_cb);
-
+#else
+  gnutls_certificate_set_retrieve_function(xcred, mailstream_gnutls_client_cert_cb);
+#endif
   gnutls_set_default_priority(session);
   gnutls_priority_set_direct(session, "NORMAL", NULL);
 
@@ -547,12 +633,14 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, time_t timeout,
   gnutls_dh_set_prime_bits(session, 512);
   
   if (timeout == 0) {
-		timeout_value = mailstream_network_delay.tv_sec * 1000 + timeout.tv_usec / 1000;
+		timeout_value = mailstream_network_delay.tv_sec * 1000 + mailstream_network_delay.tv_usec / 1000;
   }
   else {
 		timeout_value = timeout;
   }
+#if GNUTLS_VERSION_NUMBER >= 0x030100
 	gnutls_handshake_set_timeout(session, timeout_value);
+#endif
 
   do {
     r = gnutls_handshake(session);
@@ -593,11 +681,6 @@ static struct mailstream_ssl_data * ssl_data_new(int fd, time_t timeout,
  err:
   return NULL;
 }
-static struct mailstream_ssl_data * tls_data_new(int fd, time_t timeout,
-  void (* callback)(struct mailstream_ssl_context * ssl_context, void * cb_data), void * cb_data)
-{
-  return ssl_data_new(fd, timeout, callback, cb_data);
-}
 #endif
 
 static void  ssl_data_free(struct mailstream_ssl_data * ssl_data)
@@ -627,8 +710,7 @@ static void  ssl_data_close(struct mailstream_ssl_data * ssl_data)
   gnutls_deinit(ssl_data->session);
 
   MUTEX_LOCK(&ssl_lock);
-  if(!gnutls_init_not_required)
-    gnutls_global_deinit();
+  gnutls_global_deinit();
   MUTEX_UNLOCK(&ssl_lock);
 
   ssl_data->session = NULL;
@@ -650,10 +732,7 @@ static mailstream_low * mailstream_low_ssl_open_full(int fd, int starttls, time_
   mailstream_low * s;
   struct mailstream_ssl_data * ssl_data;
 
-  if (starttls)
-    ssl_data = tls_data_new(fd, timeout, callback, cb_data);
-  else
-    ssl_data = ssl_data_new(fd, timeout, callback, cb_data);
+  ssl_data = ssl_data_new(fd, timeout, callback, cb_data);
 
   if (ssl_data == NULL)
     goto err;
@@ -726,18 +805,21 @@ static int mailstream_low_ssl_get_fd(mailstream_low * s)
 
 static int wait_read(mailstream_low * s)
 {
-  fd_set fds_read;
   struct timeval timeout;
-  int fd;
+  int cancellation_fd;
   struct mailstream_ssl_data * ssl_data;
-  int max_fd;
   int r;
   int cancelled;
-  int got_data;
-#ifdef WIN32
+#if defined(WIN32)
+  fd_set fds_read;
   HANDLE event;
+#elif USE_POLL
+  struct pollfd pfd[2];
+#else
+  fd_set fds_read;
+  int max_fd;
 #endif
-  
+
   ssl_data = (struct mailstream_ssl_data *) s->data;
   if (s->timeout == 0) {
     timeout = mailstream_network_delay;
@@ -752,35 +834,46 @@ static int wait_read(mailstream_low * s)
     return 0;
 #endif
 
+  cancellation_fd = mailstream_cancel_get_fd(ssl_data->cancel);
+#if defined(WIN32)
   FD_ZERO(&fds_read);
-  fd = mailstream_cancel_get_fd(ssl_data->cancel);
-  FD_SET(fd, &fds_read);
-#ifdef WIN32
+  FD_SET(cancellation_fd, &fds_read);
   event = CreateEvent(NULL, TRUE, FALSE, NULL);
   WSAEventSelect(ssl_data->fd, event, FD_READ | FD_CLOSE);
   FD_SET(event, &fds_read);
   r = WaitForMultipleObjects(fds_read.fd_count, fds_read.fd_array, FALSE, timeout.tv_sec * 1000 + timeout.tv_usec / 1000);
   if (WAIT_TIMEOUT == r) {
-		WSAEventSelect(ssl_data->fd, event, 0);
-		CloseHandle(event);
+    WSAEventSelect(ssl_data->fd, event, 0);
+    CloseHandle(event);
     return -1;
-	}
+  }
   
-  cancelled = (fds_read.fd_array[r - WAIT_OBJECT_0] == fd);
-  got_data = (fds_read.fd_array[r - WAIT_OBJECT_0] == event);
-	WSAEventSelect(ssl_data->fd, event, 0);
-	CloseHandle(event);
-#else
-  FD_SET(ssl_data->fd, &fds_read);
-  max_fd = ssl_data->fd;
-  if (fd > max_fd)
-    max_fd = fd;
-  r = select(max_fd + 1, &fds_read, NULL, NULL, &timeout);
+  cancelled = (fds_read.fd_array[r - WAIT_OBJECT_0] == cancellation_fd);
+  WSAEventSelect(ssl_data->fd, event, 0);
+  CloseHandle(event);
+#elif USE_POLL
+  pfd[0].fd = ssl_data->fd;
+  pfd[0].events = POLLIN;
+  pfd[0].revents = 0;
+
+  pfd[1].fd = cancellation_fd;
+  pfd[1].events = POLLIN;
+  pfd[1].revents = 0;
+
+  r = poll(&pfd[0], 2, timeout.tv_sec * 1000 + timeout.tv_usec / 1000);
   if (r <= 0)
     return -1;
   
-  cancelled = (FD_ISSET(fd, &fds_read));
-  got_data = FD_ISSET(ssl_data->fd, &fds_read);
+  cancelled = pfd[1].revents & POLLIN;
+#else
+  FD_ZERO(&fds_read);
+  FD_SET(cancellation_fd, &fds_read);
+  FD_SET(ssl_data->fd, &fds_read);
+  max_fd = cancellation_fd > ssl_data->fd ? cancellation_fd : ssl_data->fd;
+  r = select(max_fd + 1, &fds_read, NULL, NULL, &timeout);
+  if (r <= 0)
+      return -1;
+  cancelled = FD_ISSET(cancellation_fd, &fds_read);
 #endif
   if (cancelled) {
     /* cancelled */
@@ -806,7 +899,7 @@ static ssize_t mailstream_low_ssl_read(mailstream_low * s,
   while (1) {
     int ssl_r;
     
-    r = SSL_read(ssl_data->ssl_conn, buf, count);
+    r = SSL_read(ssl_data->ssl_conn, buf, (int) count);
     if (r > 0)
       return r;
     
@@ -870,17 +963,22 @@ static ssize_t mailstream_low_ssl_read(mailstream_low * s,
 
 static int wait_write(mailstream_low * s)
 {
-  fd_set fds_read;
-  fd_set fds_write;
   struct timeval timeout;
   int r;
-  int fd;
+  int cancellation_fd;
   struct mailstream_ssl_data * ssl_data;
-  int max_fd;
   int cancelled;
   int write_enabled;
-#ifdef WIN32
+#if defined(WIN32)
+  fd_set fds_read;
+  fd_set fds_write;
   HANDLE event;
+#elif USE_POLL
+  struct pollfd pfd[2];
+#else
+  fd_set fds_read;
+  fd_set fds_write;
+  int max_fd;
 #endif
   
   ssl_data = (struct mailstream_ssl_data *) s->data;
@@ -895,11 +993,11 @@ static int wait_write(mailstream_low * s)
     timeout.tv_usec = 0;
   }
   
+  cancellation_fd = mailstream_cancel_get_fd(ssl_data->cancel);
+#if defined(WIN32)
   FD_ZERO(&fds_read);
-  fd = mailstream_cancel_get_fd(ssl_data->cancel);
-  FD_SET(fd, &fds_read);
   FD_ZERO(&fds_write);
-#ifdef WIN32
+  FD_SET(cancellation_fd, &fds_read);
   event = CreateEvent(NULL, TRUE, FALSE, NULL);
   WSAEventSelect(ssl_data->fd, event, FD_WRITE | FD_CLOSE);
   FD_SET(event, &fds_read);
@@ -910,22 +1008,37 @@ static int wait_write(mailstream_low * s)
     return -1;
 	}
   
-  cancelled = (fds_read.fd_array[r - WAIT_OBJECT_0] == fd) /* SEB 20070709 */;
+  cancelled = (fds_read.fd_array[r - WAIT_OBJECT_0] == cancellation_fd) /* SEB 20070709 */;
   write_enabled = (fds_read.fd_array[r - WAIT_OBJECT_0] == event);
 	WSAEventSelect(ssl_data->fd, event, 0);
 	CloseHandle(event);
+#elif USE_POLL
+  pfd[0].fd = ssl_data->fd;
+  pfd[0].events = POLLOUT;
+  pfd[0].revents = 0;
+
+  pfd[1].fd = cancellation_fd;
+  pfd[1].events = POLLIN;
+  pfd[1].revents = 0;
+
+  r = poll(&pfd[0], 2, timeout.tv_sec * 1000 + timeout.tv_usec / 1000);
+  if (r <= 0)
+    return -1;
+ 
+  cancelled = pfd[1].revents & POLLIN;
+  write_enabled = pfd[0].revents & POLLOUT;
 #else
+  FD_ZERO(&fds_read);
+  FD_ZERO(&fds_write);
+  FD_SET(cancellation_fd, &fds_read);
   FD_SET(ssl_data->fd, &fds_write);
   
-  max_fd = ssl_data->fd;
-  if (fd > max_fd)
-    max_fd = fd;
-  
+  max_fd = cancellation_fd > ssl_data->fd ? cancellation_fd : ssl_data->fd;
   r = select(max_fd + 1, &fds_read, &fds_write, NULL, &timeout);
   if (r <= 0)
     return -1;
-  
-  cancelled = FD_ISSET(fd, &fds_read);
+
+  cancelled = FD_ISSET(cancellation_fd, &fds_read);
   write_enabled = FD_ISSET(ssl_data->fd, &fds_write);
 #endif
   
@@ -954,7 +1067,7 @@ static ssize_t mailstream_low_ssl_write(mailstream_low * s,
   if (r <= 0)
     return r;
   
-  r = SSL_write(ssl_data->ssl_conn, buf, count);
+  r = SSL_write(ssl_data->ssl_conn, buf, (int) count);
   if (r > 0)
     return r;
   
@@ -1061,7 +1174,6 @@ ssize_t mailstream_ssl_get_certificate(mailstream *stream, unsigned char **cert_
   const gnutls_datum *raw_cert_list;
   unsigned int raw_cert_list_length;
   gnutls_x509_crt cert = NULL;
-  char output[10*1024];
   size_t cert_size;
 #endif
 
@@ -1081,9 +1193,13 @@ ssize_t mailstream_ssl_get_certificate(mailstream *stream, unsigned char **cert_
   if (cert == NULL)
     return -1;
   
-  *cert_DER = NULL;
-  len = (ssize_t) i2d_X509(cert, cert_DER);
-  
+  len = i2d_X509(cert, NULL);
+  * cert_DER = malloc(len);
+  if (* cert_DER == NULL)
+    return -1;
+
+  i2d_X509(cert, cert_DER);
+
 	X509_free(cert);
 
   return len;
@@ -1095,15 +1211,18 @@ ssize_t mailstream_ssl_get_certificate(mailstream *stream, unsigned char **cert_
   && gnutls_certificate_type_get(session) == GNUTLS_CRT_X509
   &&  gnutls_x509_crt_init(&cert) >= 0
   &&  gnutls_x509_crt_import(cert, &raw_cert_list[0], GNUTLS_X509_FMT_DER) >= 0) {
-    cert_size = sizeof(output);
-    if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, output, &cert_size) < 0)
+    cert_size = 0;
+    if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, NULL, &cert_size) 
+        != GNUTLS_E_SHORT_MEMORY_BUFFER)
       return -1;
-    
-    *cert_DER = malloc (cert_size + 1);
+
+    *cert_DER = malloc (cert_size);
     if (*cert_DER == NULL)
       return -1;
-    
-    memcpy (*cert_DER, output, cert_size);
+
+    if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, *cert_DER, &cert_size) < 0)
+      return -1;
+
     len = (ssize_t)cert_size;
     gnutls_x509_crt_deinit(cert);
     
@@ -1194,7 +1313,7 @@ int mailstream_ssl_set_client_certificate_data(struct mailstream_ssl_context * s
   tmp.data = x509_der;
   tmp.size = len;
   if ((r = gnutls_x509_crt_import(ssl_context->client_x509, &tmp, GNUTLS_X509_FMT_DER)) < 0) {
-    gnutls_x509_crt_deinit(ssl_context->client_x509); /* ici */
+    gnutls_x509_crt_deinit(ssl_context->client_x509);
     ssl_context->client_x509 = NULL;
     return -1;
   }
@@ -1253,6 +1372,47 @@ int mailstream_ssl_set_server_certicate(struct mailstream_ssl_context * ssl_cont
 #endif /* USE_SSL */
 }
 
+LIBETPAN_EXPORT
+int mailstream_ssl_set_server_name(struct mailstream_ssl_context * ssl_context,
+    char * hostname)
+{
+  int r = -1;
+
+#ifdef USE_SSL
+# ifdef USE_GNUTLS
+  if (hostname != NULL) {
+    r = gnutls_server_name_set(ssl_context->session, GNUTLS_NAME_DNS, hostname, strlen(hostname));
+  }
+  else {
+    r = gnutls_server_name_set(ssl_context->session, GNUTLS_NAME_DNS, "", 0U);
+  }
+# else /* !USE_GNUTLS */
+#  if (OPENSSL_VERSION_NUMBER >= 0x10000000L)
+  if (hostname != NULL) {
+    /* Unfortunately we can't set this in the openssl session yet since it
+     * hasn't been created yet; we only have the openssl context at this point.
+     * We will set it in the openssl session when we create it, soon after the
+     * client callback that we expect to be calling us (since it is the way the
+     * client gets our mailstream_ssl_context) returns (see
+     * ssl_data_new_full()) but we cannot rely on the client persisting it. We
+     * must therefore take a temporary copy here, which we free once we've set
+     * it in the openssl session. */
+    ssl_context->server_name = strdup(hostname);
+  }
+  else {
+    if (ssl_context->server_name != NULL) {
+      free(ssl_context->server_name);
+    }
+    ssl_context->server_name = NULL;
+  }
+  r = 0;
+#  endif /* (OPENSSL_VERSION_NUMBER >= 0x10000000L) */
+# endif /* !USE_GNUTLS */
+#endif /* USE_SSL */
+
+  return r;
+}
+
 #ifdef USE_SSL
 #ifndef USE_GNUTLS
 static struct mailstream_ssl_context * mailstream_ssl_context_new(SSL_CTX * open_ssl_ctx, int fd)
@@ -1266,6 +1426,9 @@ static struct mailstream_ssl_context * mailstream_ssl_context_new(SSL_CTX * open
   ssl_ctx->openssl_ssl_ctx = open_ssl_ctx;
   ssl_ctx->client_x509 = NULL;
   ssl_ctx->client_pkey = NULL;
+#if (OPENSSL_VERSION_NUMBER >= 0x10000000L)
+  ssl_ctx->server_name = NULL;
+#endif /* (OPENSSL_VERSION_NUMBER >= 0x10000000L) */
   ssl_ctx->fd = fd;
   
   return ssl_ctx;
@@ -1273,8 +1436,14 @@ static struct mailstream_ssl_context * mailstream_ssl_context_new(SSL_CTX * open
 
 static void mailstream_ssl_context_free(struct mailstream_ssl_context * ssl_ctx)
 {
-  if (ssl_ctx)
+  if (ssl_ctx != NULL) {
+#if (OPENSSL_VERSION_NUMBER >= 0x10000000L)
+    if (ssl_ctx->server_name != NULL) {
+      free(ssl_ctx->server_name);
+    }
+#endif /* (OPENSSL_VERSION_NUMBER >= 0x10000000L) */
     free(ssl_ctx);
+  }
 }
 #else
 static struct mailstream_ssl_context * mailstream_ssl_context_new(gnutls_session session, int fd)
@@ -1332,3 +1501,77 @@ static struct mailstream_cancel * mailstream_low_ssl_get_cancel(mailstream_low *
   return NULL;
 #endif
 }
+
+carray * mailstream_low_ssl_get_certificate_chain(mailstream_low * s)
+{
+#ifdef USE_SSL
+  struct mailstream_ssl_data * ssl_data;
+  carray * result;
+  int skpos;
+#ifndef USE_GNUTLS
+  STACK_OF(X509) * skx;
+  
+  ssl_data = (struct mailstream_ssl_data *) s->data;
+  if (!(skx = SSL_get_peer_cert_chain(ssl_data->ssl_conn))) {
+    return NULL;
+  }
+  
+  result = carray_new(4);
+  for(skpos = 0 ; skpos < sk_num((_STACK *) skx) ; skpos ++) {
+    X509 * x = (X509 *) sk_value((_STACK *) skx, skpos);
+    unsigned char * p;
+    MMAPString * str;
+    int length = i2d_X509(x, NULL);
+    str = mmap_string_sized_new(length);
+    p = (unsigned char *) str->str;
+    str->len = length;
+    i2d_X509(x, &p);
+    carray_add(result, str, NULL);
+  }
+  
+  return result;
+#else
+  gnutls_session session = NULL;
+  const gnutls_datum *raw_cert_list;
+  unsigned int raw_cert_list_length;
+
+  ssl_data = (struct mailstream_ssl_data *) s->data;
+
+  session = ssl_data->session;
+  raw_cert_list = gnutls_certificate_get_peers(session, &raw_cert_list_length);
+
+  if (raw_cert_list && gnutls_certificate_type_get(session) == GNUTLS_CRT_X509) {
+    result = carray_new(4);
+    for(skpos = 0 ; skpos < raw_cert_list_length ; skpos ++) {
+      gnutls_x509_crt cert = NULL;
+      if (gnutls_x509_crt_init(&cert) >= 0
+       && gnutls_x509_crt_import(cert, &raw_cert_list[skpos], GNUTLS_X509_FMT_DER) >= 0) {
+         size_t cert_size = 0;
+         MMAPString * str = NULL;
+         unsigned char * p;
+
+         if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, NULL, &cert_size)
+	     == GNUTLS_E_SHORT_MEMORY_BUFFER) {
+           str = mmap_string_sized_new(cert_size);
+           p = (unsigned char *) str->str;
+           str->len = cert_size;
+	 }
+	 if (str != NULL &&
+             gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, p, &cert_size) >= 0) {
+           carray_add(result, str, NULL);
+	 } else {
+	   return NULL;
+	 }
+         gnutls_x509_crt_deinit(cert);
+       }
+    }
+  }
+
+  return result;
+
+  return NULL;
+#endif
+#else
+  return NULL;
+#endif
+}
