@@ -29,8 +29,12 @@
  * SUCH DAMAGE.
  */
 
+#ifdef HAVE_CONFIG_H
+#	include <config.h>
+#endif
+
 /*
- * $Id: md5.c,v 1.8 2005/06/01 12:21:57 smarinier Exp $
+ * $Id: md5.c,v 1.11 2008/02/20 22:15:50 hoa Exp $
  */
 
 /* MD5C.C - RSA Data Security, Inc., MD5 message-digest algorithm
@@ -60,7 +64,9 @@ documentation and/or software.
 
 /* do i need all of this just for htonl()? damn. */
 #include <sys/types.h>
-#ifndef _MSC_VER
+#ifdef WIN32
+#	include <win_etpan.h>
+#else
 #	include <sys/param.h>
 #	include <sys/socket.h>
 #	include <netinet/in.h>
@@ -90,12 +96,12 @@ documentation and/or software.
 #define S43 15
 #define S44 21
 
-static void MD5Transform PROTO_LIST ((UINT4 [4], unsigned char [64]));
+static void MD5Transform PROTO_LIST ((UINT4 [4], const unsigned char [64]));
 static void Encode PROTO_LIST
        ((unsigned char *, UINT4 *, unsigned int)); 
 static void Decode PROTO_LIST
-       ((UINT4 *, unsigned char *, unsigned int)); 
-static void MD5_memcpy PROTO_LIST ((POINTER, POINTER, unsigned int));
+       ((UINT4 *, const unsigned char *, unsigned int)); 
+static void MD5_memcpy PROTO_LIST ((POINTER, CONST_POINTER, unsigned int));
 static void MD5_memset PROTO_LIST ((POINTER, int, unsigned int));
 
 static unsigned char PADDING[64] = {
@@ -154,13 +160,13 @@ MD5_CTX *context; /* context */
 
 void MD5Update (context, input, inputLen)
 MD5_CTX *context; /* context */
-unsigned char *input; /* input block */
+const unsigned char *input; /* input block */
 unsigned int inputLen; /* length of input block */
 {
-       unsigned int i, index, partLen; 
+       unsigned int i, indx, partLen; 
 
          /* Compute number of bytes mod 64 */
-         index = (unsigned int)((context->count[0] >> 3) & 0x3F);
+         indx = (unsigned int)((context->count[0] >> 3) & 0x3F);
 
          /* Update number of bits */
          if ((context->count[0] += ((UINT4)inputLen << 3))
@@ -168,27 +174,27 @@ unsigned int inputLen; /* length of input block */
         context->count[1]++;
          context->count[1] += ((UINT4)inputLen >> 29);
 
-       partLen = 64 - index; 
+       partLen = 64 - indx; 
 
          /* Transform as many times as possible.
 
 */
        if (inputLen >= partLen) { 
        MD5_memcpy 
-       ((POINTER)&context->buffer[index], (POINTER)input, partLen); MD5Transform
+       ((POINTER)&context->buffer[indx], (POINTER)input, partLen); MD5Transform
        (context->state, context->buffer); 
 
        for (i = partLen; i + 63 < inputLen; i += 64) 
        MD5Transform (context->state, &input[i]); 
 
-       index = 0; 
+       indx = 0; 
        } 
        else 
        i = 0; 
 
          /* Buffer remaining input */
          MD5_memcpy
-        ((POINTER)&context->buffer[index], (POINTER)&input[i],
+        ((POINTER)&context->buffer[indx], (POINTER)&input[i],
          inputLen-i);
 
 }
@@ -203,7 +209,7 @@ unsigned char digest[16]; /* message digest */
 MD5_CTX *context; /* context */
 {
        unsigned char bits[8]; 
-       unsigned int index, padLen; 
+       unsigned int indx, padLen; 
 
          /* Save number of bits */
          Encode (bits, context->count, 8);
@@ -211,8 +217,8 @@ MD5_CTX *context; /* context */
          /* Pad out to 56 mod 64.
 
 */
-       index = (unsigned int)((context->count[0] >> 3) & 0x3f); 
-       padLen = (index < 56) ? (56 - index) : (120 - index); 
+       indx = (unsigned int)((context->count[0] >> 3) & 0x3f); 
+       padLen = (indx < 56) ? (56 - indx) : (120 - indx); 
        MD5Update (context, PADDING, padLen); 
 
          /* Append length (before padding) */
@@ -233,7 +239,7 @@ MD5_CTX *context; /* context */
 
 static void MD5Transform (state, block)
 UINT4 state[4];
-unsigned char block[64];
+const unsigned char block[64];
 {
        UINT4 a = state[0], b = state[1], c = state[2], d = state[3], x[16]; 
 
@@ -348,7 +354,7 @@ unsigned int len;
 
 static void Decode (output, input, len)
 UINT4 *output;
-unsigned char *input;
+const unsigned char *input;
 unsigned int len;
 {
        unsigned int i, j; 
@@ -364,7 +370,7 @@ unsigned int len;
 
 static void MD5_memcpy (output, input, len)
 POINTER output;
-POINTER input;
+CONST_POINTER input;
 unsigned int len;
 {
        unsigned int i; 
@@ -442,9 +448,9 @@ void hmac_md5_init(HMAC_MD5_CTX *hmac,
   MD5Update(&hmac->octx, k_opad, 64);     /* apply outer pad */
 
   /* scrub the pads and key context (if used) */
-  MD5_memset(&k_ipad, 0, sizeof(k_ipad));
-  MD5_memset(&k_opad, 0, sizeof(k_opad));
-  MD5_memset(&tk, 0, sizeof(tk));
+  MD5_memset(&k_ipad[0], 0, sizeof(k_ipad));
+  MD5_memset(&k_opad[0], 0, sizeof(k_opad));
+  MD5_memset(&tk[0], 0, sizeof(tk));
 
   /* and we're done. */
 }
@@ -469,7 +475,7 @@ void hmac_md5_precalc(HMAC_MD5_STATE *state,
     state->istate[lupe] = htonl(hmac.ictx.state[lupe]);
     state->ostate[lupe] = htonl(hmac.octx.state[lupe]);
   }
-  MD5_memset(&hmac, 0, sizeof(hmac));
+  MD5_memset((POINTER) &hmac, 0, sizeof(hmac));
 }
 
 
@@ -477,7 +483,7 @@ void hmac_md5_import(HMAC_MD5_CTX *hmac,
 		     HMAC_MD5_STATE *state)
 {
   unsigned lupe;
-  MD5_memset(hmac, 0, sizeof(HMAC_MD5_CTX));
+  MD5_memset((POINTER) hmac, 0, sizeof(HMAC_MD5_CTX));
   for (lupe = 0; lupe < 4; lupe++) {
     hmac->ictx.state[lupe] = ntohl(state->istate[lupe]);
     hmac->octx.state[lupe] = ntohl(state->ostate[lupe]);
