@@ -30,7 +30,7 @@
  */
 
 /*
- * $Id: mailstorage_tools.h,v 1.7 2005/08/15 11:08:41 hoa Exp $
+ * $Id: mailstorage_tools.h,v 1.8 2007/06/30 12:58:21 hoa Exp $
  */
 
 #include "mailstorage.h"
@@ -54,6 +54,19 @@ int mailstorage_generic_connect(mailsession_driver * driver,
     char * flags_directory,
     mailsession ** result);
 
+int mailstorage_generic_connect_with_local_address(mailsession_driver * driver,
+    char * servername,
+    uint16_t port,
+    char * local_address,
+    uint16_t local_port,
+    char * command,
+    int connection_type,
+    int cache_function_id,
+    char * cache_directory,
+    int flags_function_id,
+    char * flags_directory,
+    mailsession ** result);
+
 int mailstorage_generic_auth(mailsession * session,
     int connect_result,
     int auth_type,
