@@ -117,6 +117,9 @@
 http://www.ietf.org/ids.by.wg/imapext.html
 */
 
+static inline void imap_logger(mailstream * s, int log_type,
+    const char * str, size_t size, void * context);
+
 static int parse_greeting(mailimap * session,
 			   struct mailimap_greeting ** result);
 
@@ -534,6 +537,7 @@ int mailimap_connect(mailimap * session, mailstream * s)
     return MAILIMAP_ERROR_BAD_STATE;
 
   session->imap_stream = s;
+  mailstream_set_logger(s, imap_logger, session);
   
   if (session->imap_connection_info)
     mailimap_connection_info_free(session->imap_connection_info);
@@ -612,7 +616,7 @@ int mailimap_append(mailimap * session, const char * mailbox,
   indx = 0;
 
   r = mailimap_continue_req_parse(session->imap_stream,
-      session->imap_stream_buffer,
+      session->imap_stream_buffer, NULL,
       &indx, &cont_req,
       session->imap_progr_rate, session->imap_progr_fun);
   if (r == MAILIMAP_NO_ERROR)
@@ -1141,6 +1145,98 @@ int mailimap_uid_copy(mailimap * session, struct mailimap_set * set,
   }
 }
 
+LIBETPAN_EXPORT
+int mailimap_move(mailimap * session, struct mailimap_set * set,
+                  const char * mb)
+{
+  struct mailimap_response * response;
+  int r;
+  int error_code;
+
+  if (session->imap_state != MAILIMAP_STATE_SELECTED)
+    return MAILIMAP_ERROR_BAD_STATE;
+
+  r = mailimap_send_current_tag(session);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  r = mailimap_move_send(session->imap_stream, set, mb);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  r = mailimap_crlf_send(session->imap_stream);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  if (mailstream_flush(session->imap_stream) == -1)
+    return MAILIMAP_ERROR_STREAM;
+
+  if (mailimap_read_line(session) == NULL)
+    return MAILIMAP_ERROR_STREAM;
+
+  r = mailimap_parse_response(session, &response);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  error_code = response->rsp_resp_done->rsp_data.rsp_tagged->rsp_cond_state->rsp_type;
+
+  mailimap_response_free(response);
+
+  switch (error_code) {
+    case MAILIMAP_RESP_COND_STATE_OK:
+      return MAILIMAP_NO_ERROR;
+
+    default:
+      return MAILIMAP_ERROR_MOVE;
+  }
+}
+
+LIBETPAN_EXPORT
+int mailimap_uid_move(mailimap * session, struct mailimap_set * set,
+                      const char * mb)
+{
+  struct mailimap_response * response;
+  int r;
+  int error_code;
+
+  if (session->imap_state != MAILIMAP_STATE_SELECTED)
+    return MAILIMAP_ERROR_BAD_STATE;
+
+  r = mailimap_send_current_tag(session);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  r = mailimap_uid_move_send(session->imap_stream, set, mb);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  r = mailimap_crlf_send(session->imap_stream);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  if (mailstream_flush(session->imap_stream) == -1)
+    return MAILIMAP_ERROR_STREAM;
+
+  if (mailimap_read_line(session) == NULL)
+    return MAILIMAP_ERROR_STREAM;
+
+  r = mailimap_parse_response(session, &response);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+
+  error_code = response->rsp_resp_done->rsp_data.rsp_tagged->rsp_cond_state->rsp_type;
+
+  mailimap_response_free(response);
+
+  switch (error_code) {
+    case MAILIMAP_RESP_COND_STATE_OK:
+      return MAILIMAP_NO_ERROR;
+
+    default:
+      return MAILIMAP_ERROR_UID_MOVE;
+  }
+}
+
 LIBETPAN_EXPORT
 int mailimap_create(mailimap * session, const char * mb)
 {
@@ -1547,14 +1643,14 @@ static int sasl_getsimple(void * context, int id,
     if (result != NULL)
       * result = session->imap_sasl.sasl_login;
     if (len != NULL)
-      * len = strlen(session->imap_sasl.sasl_login);
+      * len = (unsigned) strlen(session->imap_sasl.sasl_login);
     return SASL_OK;
     
   case SASL_CB_AUTHNAME:
     if (result != NULL)
       * result = session->imap_sasl.sasl_auth_name;
     if (len != NULL)
-      * len = strlen(session->imap_sasl.sasl_auth_name);
+      * len = (unsigned) strlen(session->imap_sasl.sasl_auth_name);
     return SASL_OK;
   }
   
@@ -1719,7 +1815,7 @@ int mailimap_authenticate(mailimap * session, const char * auth_type,
     indx = 0;
     
     r = mailimap_continue_req_parse(session->imap_stream,
-        session->imap_stream_buffer,
+        session->imap_stream_buffer, NULL,
         &indx, &cont_req,
         session->imap_progr_rate, session->imap_progr_fun);
     if (r != MAILIMAP_NO_ERROR)
@@ -1737,12 +1833,12 @@ int mailimap_authenticate(mailimap * session, const char * auth_type,
     }
     
     if (got_response) {
-      size_t response_len;
+      unsigned int response_len;
       char * decoded;
       unsigned int decoded_len;
       unsigned int max_decoded;
       
-      response_len = strlen(response_base64);
+      response_len = (unsigned int) strlen(response_base64);
       max_decoded = response_len * 3 / 4;
       decoded = malloc(max_decoded + 1);
       if (decoded == NULL) {
@@ -1851,7 +1947,6 @@ int mailimap_authenticate(mailimap * session, const char * auth_type,
 #endif
 }
 
-
 LIBETPAN_EXPORT
 int mailimap_lsub(mailimap * session, const char * mb,
     const char * list_mb, clist ** result)
@@ -1961,53 +2056,7 @@ int
 mailimap_search(mailimap * session, const char * charset,
     struct mailimap_search_key * key, clist ** result)
 {
-#if 0
-  struct mailimap_response * response;
-  int r;
-  int error_code;
-
-  if (session->imap_state != MAILIMAP_STATE_SELECTED)
-    return MAILIMAP_ERROR_BAD_STATE;
-
-  r = mailimap_send_current_tag(session);
-  if (r != MAILIMAP_NO_ERROR)
-	return r;
-
-  r = mailimap_search_send(session->imap_stream, charset, key);
-  if (r != MAILIMAP_NO_ERROR)
-	return r;
-
-  r = mailimap_crlf_send(session->imap_stream);
-  if (r != MAILIMAP_NO_ERROR)
-	return r;
-
-  if (mailstream_flush(session->imap_stream) == -1)
-    return MAILIMAP_ERROR_STREAM;
-
-  if (mailimap_read_line(session) == NULL)
-    return MAILIMAP_ERROR_STREAM;
-
-  r = mailimap_parse_response(session, &response);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  * result = session->imap_response_info->rsp_search_result;
-  session->imap_response_info->rsp_search_result = NULL;
-
-  error_code = response->rsp_resp_done->rsp_data.rsp_tagged->rsp_cond_state->rsp_type;
-
-  mailimap_response_free(response);
-
-  switch (error_code) {
-  case MAILIMAP_RESP_COND_STATE_OK:
-    return MAILIMAP_NO_ERROR;
-
-  default:
-    return MAILIMAP_ERROR_SEARCH;
-  }
-#else
   return mailimap_search_modseq(session, charset, key, result, NULL);
-#endif
 }
 
 LIBETPAN_EXPORT
@@ -2015,53 +2064,19 @@ int
 mailimap_uid_search(mailimap * session, const char * charset,
     struct mailimap_search_key * key, clist ** result)
 {
-#if 0
-  struct mailimap_response * response;
-  int r;
-  int error_code;
-
-  if (session->imap_state != MAILIMAP_STATE_SELECTED)
-    return MAILIMAP_ERROR_BAD_STATE;
-
-  r = mailimap_send_current_tag(session);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  r = mailimap_uid_search_send(session->imap_stream, charset, key);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  r = mailimap_crlf_send(session->imap_stream);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  if (mailstream_flush(session->imap_stream) == -1)
-    return MAILIMAP_ERROR_STREAM;
-
-  if (mailimap_read_line(session) == NULL)
-    return MAILIMAP_ERROR_STREAM;
-
-  r = mailimap_parse_response(session, &response);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  * result = session->imap_response_info->rsp_search_result;
-  session->imap_response_info->rsp_search_result = NULL;
-
-  error_code = response->rsp_resp_done->rsp_data.rsp_tagged->rsp_cond_state->rsp_type;
-
-  mailimap_response_free(response);
-
-  switch (error_code) {
-  case MAILIMAP_RESP_COND_STATE_OK:
-    return MAILIMAP_NO_ERROR;
-
-  default:
-    return MAILIMAP_ERROR_UID_SEARCH;
-  }
-#else
   return mailimap_uid_search_modseq(session, charset, key, result, NULL);
-#endif
+}
+
+LIBETPAN_EXPORT int mailimap_search_literalplus(mailimap * session, const char * charset,
+                                                struct mailimap_search_key * key, clist ** result)
+{
+  return mailimap_search_literalplus_modseq(session, charset, key, result, NULL);
+}
+
+LIBETPAN_EXPORT int mailimap_uid_search_literalplus(mailimap * session, const char * charset,
+                                                    struct mailimap_search_key * key, clist ** result)
+{
+  return mailimap_uid_search_literalplus_modseq(session, charset, key, result, NULL);
 }
 
 LIBETPAN_EXPORT
@@ -2071,62 +2086,6 @@ void mailimap_search_result_free(clist * search_result)
   clist_free(search_result);
 }
 
-#if 0
-LIBETPAN_EXPORT
-int
-mailimap_select(mailimap * session, const char * mb)
-{
-  struct mailimap_response * response;
-  int r;
-  int error_code;
-
-  if ((session->imap_state != MAILIMAP_STATE_AUTHENTICATED) &&
-      (session->imap_state != MAILIMAP_STATE_SELECTED))
-    return MAILIMAP_ERROR_BAD_STATE;
-
-  r = mailimap_send_current_tag(session);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  r = mailimap_select_send(session->imap_stream, mb);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-  
-  r = mailimap_crlf_send(session->imap_stream);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  if (mailstream_flush(session->imap_stream) == -1)
-    return MAILIMAP_ERROR_STREAM;
-
-  if (mailimap_read_line(session) == NULL)
-    return MAILIMAP_ERROR_STREAM;
-
-  if (session->imap_selection_info != NULL)
-    mailimap_selection_info_free(session->imap_selection_info);
-  session->imap_selection_info = mailimap_selection_info_new();
-
-  r = mailimap_parse_response(session, &response);
-  if (r != MAILIMAP_NO_ERROR)
-    return r;
-
-  error_code = response->rsp_resp_done->rsp_data.rsp_tagged->rsp_cond_state->rsp_type;
-
-  mailimap_response_free(response);
-
-  switch (error_code) {
-  case MAILIMAP_RESP_COND_STATE_OK:
-    session->imap_state = MAILIMAP_STATE_SELECTED;
-    return MAILIMAP_NO_ERROR;
-
-  default:
-    mailimap_selection_info_free(session->imap_selection_info);
-    session->imap_selection_info = NULL;
-    session->imap_state = MAILIMAP_STATE_AUTHENTICATED;
-    return MAILIMAP_ERROR_SELECT;
-  }
-}
-#else
 LIBETPAN_EXPORT
 int
 mailimap_select(mailimap * session, const char * mb)
@@ -2134,7 +2093,49 @@ mailimap_select(mailimap * session, const char * mb)
 	uint64_t dummy;
 	return mailimap_select_condstore_optional(session, mb, 0, &dummy);
 }
-#endif
+
+LIBETPAN_EXPORT
+int
+mailimap_custom_command(mailimap * session, const char * command)
+{
+  int r;
+  int error_code;
+  struct mailimap_response * response;
+  
+  r = mailimap_send_current_tag(session);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+  
+  r = mailimap_send_custom_command(session->imap_stream, command);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+  
+  r = mailimap_crlf_send(session->imap_stream);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+  
+  if (mailstream_flush(session->imap_stream) == -1)
+    return MAILIMAP_ERROR_STREAM;
+  
+  if (mailimap_read_line(session) == NULL)
+    return MAILIMAP_ERROR_STREAM;
+  
+  r = mailimap_parse_response(session, &response);
+  if (r != MAILIMAP_NO_ERROR)
+    return r;
+  
+  error_code = response->rsp_resp_done->rsp_data.rsp_tagged->rsp_cond_state->rsp_type;
+  
+  mailimap_response_free(response);
+  
+  switch (error_code) {
+    case MAILIMAP_RESP_COND_STATE_OK:
+      return MAILIMAP_NO_ERROR;
+      
+    default:
+      return MAILIMAP_ERROR_CUSTOM_COMMAND;
+  }
+}
 
 LIBETPAN_EXPORT
 int
@@ -2443,7 +2444,12 @@ int mailimap_send_current_tag(mailimap * session)
   int r;
   
   session->imap_tag ++;
-  snprintf(tag_str, 15, "%i", session->imap_tag);
+  
+  if(mailimap_is_163_workaround_enabled(session))
+    snprintf(tag_str, 15, "C%i", session->imap_tag);
+  else
+    snprintf(tag_str, 15, "%i", session->imap_tag);	
+  
 
   r = mailimap_tag_send(session->imap_stream, tag_str);
   if (r != MAILIMAP_NO_ERROR)
@@ -2461,6 +2467,7 @@ int mailimap_parse_response(mailimap * session,
 {
   size_t indx;
   struct mailimap_response * response;
+  struct mailimap_parser_context * parser_ctx;
   char tag_str[15];
   int r;
   
@@ -2476,10 +2483,15 @@ int mailimap_parse_response(mailimap * session,
 		session->imap_stream_buffer = buffer;
   }
 
+  parser_ctx = mailimap_parser_context_new(session);
+  if (parser_ctx == NULL)
+    return MAILIMAP_ERROR_MEMORY;
+
   if ((session->imap_body_progress_fun != NULL) ||
       (session->imap_items_progress_fun != NULL)) {
     r = mailimap_response_parse_with_context(session->imap_stream,
                                              session->imap_stream_buffer,
+                                             parser_ctx,
                                              &indx, &response,
                                              session->imap_body_progress_fun,
                                              session->imap_items_progress_fun,
@@ -2490,9 +2502,13 @@ int mailimap_parse_response(mailimap * session,
   else {
     r = mailimap_response_parse(session->imap_stream,
                                 session->imap_stream_buffer,
+                                parser_ctx,
                                 &indx, &response,
                                 session->imap_progr_rate, session->imap_progr_fun);
   }
+
+  mailimap_parser_context_free(parser_ctx);
+
   if (r != MAILIMAP_NO_ERROR)
     return r;
 
@@ -2518,7 +2534,11 @@ int mailimap_parse_response(mailimap * session,
     return MAILIMAP_ERROR_FATAL;
   }
 
-  snprintf(tag_str, 15, "%i", session->imap_tag);
+  if(mailimap_is_163_workaround_enabled(session))
+    snprintf(tag_str, 15, "C%i", session->imap_tag);
+  else
+    snprintf(tag_str, 15, "%i", session->imap_tag);
+    
   if (strcmp(response->rsp_resp_done->rsp_data.rsp_tagged->rsp_tag, tag_str) != 0) {
     mailimap_response_free(response);
     return MAILIMAP_ERROR_PROTOCOL;
@@ -2548,7 +2568,7 @@ static int parse_greeting(mailimap * session,
   session->imap_response = NULL;
 
   r = mailimap_greeting_parse(session->imap_stream,
-      session->imap_stream_buffer,
+      session->imap_stream_buffer, NULL,
       &indx, &greeting, session->imap_progr_rate,
       session->imap_progr_fun);
   if (r != MAILIMAP_NO_ERROR)
@@ -2631,9 +2651,17 @@ mailimap * mailimap_new(size_t imap_progr_rate,
     
   f->imap_msg_att_handler = NULL;
   f->imap_msg_att_handler_context = NULL;
-  
+
+  f->imap_msg_body_handler = NULL;
+  f->imap_msg_body_handler_context = NULL;
+
 	f->imap_timeout = 0;
 
+  f->imap_logger = NULL;
+  f->imap_logger_context = NULL;
+  f->is_163_workaround_enabled = 0;
+  f->is_rambler_workaround_enabled = 0;
+  f->is_qip_workaround_enabled = 0;
   return f;
   
  free_stream_buffer:
@@ -2698,8 +2726,68 @@ void mailimap_set_msg_att_handler(mailimap * session,
                                   mailimap_msg_att_handler * handler,
                                   void * context)
 {
-    session->imap_msg_att_handler = handler;
-    session->imap_msg_att_handler_context = context;
+  session->imap_msg_att_handler = handler;
+  session->imap_msg_att_handler_context = context;
 }
 
+LIBETPAN_EXPORT
+void mailimap_set_msg_body_handler(mailimap * session,
+                                   mailimap_msg_body_handler * handler,
+                                   void * context)
+{
+  session->imap_msg_body_handler = handler;
+  session->imap_msg_body_handler_context = context;
+}
 
+static inline void imap_logger(mailstream * s, int log_type,
+    const char * str, size_t size, void * context)
+{
+  mailimap * session;
+
+  session = context;
+  if (session->imap_logger == NULL)
+    return;
+
+  session->imap_logger(session, log_type, str, size, session->imap_logger_context);
+}
+
+LIBETPAN_EXPORT
+void mailimap_set_logger(mailimap * session, void (* logger)(mailimap * session, int log_type,
+    const char * str, size_t size, void * context), void * logger_context)
+{
+  session->imap_logger = logger;
+  session->imap_logger_context = logger_context;
+}
+
+LIBETPAN_EXPORT    
+void mailimap_set_163_workaround_enabled(mailimap * session, int enabled) {
+	session->is_163_workaround_enabled = enabled;
+}
+
+LIBETPAN_EXPORT
+int mailimap_is_163_workaround_enabled(mailimap * session) {
+	return session->is_163_workaround_enabled;
+}
+
+LIBETPAN_EXPORT
+void mailimap_set_rambler_workaround_enabled(mailimap * session, int enabled) {
+  if (session)
+    session->is_rambler_workaround_enabled = enabled;
+}
+
+LIBETPAN_EXPORT
+int mailimap_is_rambler_workaround_enabled(mailimap * session) {
+  return session->is_rambler_workaround_enabled;
+}
+
+LIBETPAN_EXPORT
+void mailimap_set_qip_workaround_enabled(mailimap * session, int enabled) {
+  if (session) {
+    session->is_qip_workaround_enabled = enabled;
+  }
+}
+
+LIBETPAN_EXPORT
+int mailimap_is_qip_workaround_enabled(mailimap * session) {
+  return session->is_qip_workaround_enabled;
+}
