@@ -1,3 +1,33 @@
+/*
+ * libEtPan! -- a mail stuff library
+ *
+ * Copyright (C) 2001, 2014 - DINH Viet Hoa
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. Neither the name of the libEtPan! project nor the names of its
+ *    contributors may be used to endorse or promote products derived
+ *    from this software without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
 #include "uidplus_parser.h"
 
 #include <stdio.h>
@@ -9,15 +39,15 @@
 #include "uidplus_types.h"
 #include "uidplus.h"
 
-static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index, struct mailimap_set_item ** result);
+static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx, struct mailimap_set_item ** result);
 
-static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index,
+static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx,
     struct mailimap_extension_data ** result);
 
 int mailimap_uidplus_parse(int calling_parser, mailstream * fd,
-    MMAPString * buffer, size_t * index,
+    MMAPString * buffer, struct mailimap_parser_context * parser_ctx, size_t * indx,
     struct mailimap_extension_data ** result,
     size_t progr_rate,
     progress_function * progr_fun)
@@ -25,7 +55,7 @@ int mailimap_uidplus_parse(int calling_parser, mailstream * fd,
   if (calling_parser != MAILIMAP_EXTENDED_PARSER_RESP_TEXT_CODE)
     return MAILIMAP_ERROR_PARSE;
   
-  return mailimap_uidplus_resp_code_parse(fd, buffer, index, result);
+  return mailimap_uidplus_resp_code_parse(fd, buffer, parser_ctx, indx, result);
 }
 
 
@@ -33,8 +63,8 @@ int mailimap_uidplus_parse(int calling_parser, mailstream * fd,
    uid-set         = (uniqueid / uid-range) *("," uid-set)
 */
 
-static int uid_set_item_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index, struct mailimap_set_item ** result,
+static int uid_set_item_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx, struct mailimap_set_item ** result,
     size_t progr_rate,
     progress_function * progr_fun)
 {
@@ -43,26 +73,26 @@ static int uid_set_item_parse(mailstream * fd, MMAPString * buffer,
   uint32_t uniqueid;
   size_t cur_token;
   
-  cur_token = * index;
+  cur_token = * indx;
   
-  r = mailimap_uid_range_parse(fd, buffer, &cur_token, &set_item);
+  r = mailimap_uid_range_parse(fd, buffer, parser_ctx, &cur_token, &set_item);
   if (r == MAILIMAP_NO_ERROR) {
     * result = set_item;
-    * index = cur_token;
+    * indx = cur_token;
     return r;
   }
   
   if (r != MAILIMAP_ERROR_PARSE)
     return r;
   
-  r = mailimap_uniqueid_parse(fd, buffer, &cur_token, &uniqueid);
+  r = mailimap_uniqueid_parse(fd, buffer, parser_ctx, &cur_token, &uniqueid);
   if (r == MAILIMAP_NO_ERROR) {
     set_item = mailimap_set_item_new(uniqueid, uniqueid);
     if (set_item == NULL)
       return MAILIMAP_ERROR_MEMORY;
     
     * result = set_item;
-    * index = cur_token;
+    * indx = cur_token;
     
     return MAILIMAP_NO_ERROR;
   }
@@ -75,8 +105,8 @@ static void uid_set_item_destructor(struct mailimap_set_item * set_item)
   mailimap_set_item_free(set_item);
 }
 
-static int mailimap_uid_set_parse(mailstream * fd, MMAPString *buffer,
-    size_t * index,
+static int mailimap_uid_set_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx,
     struct mailimap_set ** result)
 {
   int r;
@@ -84,9 +114,9 @@ static int mailimap_uid_set_parse(mailstream * fd, MMAPString *buffer,
   struct mailimap_set * set;
   size_t cur_token;
   
-  cur_token = * index;
+  cur_token = * indx;
   
-  r = mailimap_struct_list_parse(fd, buffer, &cur_token, &list,
+  r = mailimap_struct_list_parse(fd, buffer, parser_ctx, &cur_token, &list,
       ',',
       (mailimap_struct_parser *) uid_set_item_parse,
       (mailimap_struct_destructor *) uid_set_item_destructor,
@@ -109,7 +139,7 @@ static int mailimap_uid_set_parse(mailstream * fd, MMAPString *buffer,
   }
   
   * result = set;
-  * index = cur_token;
+  * indx = cur_token;
   
   return MAILIMAP_NO_ERROR;
 }
@@ -121,8 +151,8 @@ static int mailimap_uid_set_parse(mailstream * fd, MMAPString *buffer,
                      ; Example: 2:4 and 4:2 are equivalent.
 */
 
-static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index, struct mailimap_set_item ** result)
+static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx, struct mailimap_set_item ** result)
 {
   uint32_t first;
   uint32_t last;
@@ -130,17 +160,17 @@ static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer,
   struct mailimap_set_item * item;
   size_t cur_token;
   
-  cur_token = * index;
+  cur_token = * indx;
   
-  r = mailimap_uniqueid_parse(fd, buffer, &cur_token, &first);
+  r = mailimap_uniqueid_parse(fd, buffer, parser_ctx, &cur_token, &first);
   if (r != MAILIMAP_NO_ERROR)
     return r;
   
-  r = mailimap_colon_parse(fd, buffer, &cur_token);
+  r = mailimap_colon_parse(fd, buffer, parser_ctx, &cur_token);
   if (r != MAILIMAP_NO_ERROR)
     return r;
   
-  r = mailimap_uniqueid_parse(fd, buffer, &cur_token, &last);
+  r = mailimap_uniqueid_parse(fd, buffer, parser_ctx, &cur_token, &last);
   if (r != MAILIMAP_NO_ERROR)
     return r;
   
@@ -148,7 +178,7 @@ static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer,
   if (item == NULL)
     return MAILIMAP_ERROR_MEMORY;
   
-  * index = cur_token;
+  * indx = cur_token;
   * result = item;
   
   return MAILIMAP_NO_ERROR;
@@ -162,19 +192,19 @@ static int mailimap_uid_range_parse(mailstream * fd, MMAPString * buffer,
                      ; to append multiple messages.
 */
 
-static int mailimap_append_uid_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index,
+static int mailimap_append_uid_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx,
     struct mailimap_set ** result)
 {
-  return mailimap_uid_set_parse(fd, buffer, index, result);
+  return mailimap_uid_set_parse(fd, buffer, parser_ctx, indx, result);
 }
 
 /*
    resp-code-apnd  = "APPENDUID" SP nz-number SP append-uid
 */
 
-static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index,
+static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx,
     struct mailimap_uidplus_resp_code_apnd ** result)
 {
   int r;
@@ -183,7 +213,7 @@ static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer,
   struct mailimap_set * set;
   struct mailimap_uidplus_resp_code_apnd * resp_code_apnd;
   
-  cur_token = * index;
+  cur_token = * indx;
   
   r = mailimap_token_case_insensitive_parse(fd, buffer, &cur_token,
       "APPENDUID");
@@ -194,7 +224,7 @@ static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer,
   if (r != MAILIMAP_NO_ERROR)
     return r;
   
-  r = mailimap_nz_number_parse(fd, buffer, &cur_token, &uidvalidity);
+  r = mailimap_nz_number_parse(fd, buffer, parser_ctx, &cur_token, &uidvalidity);
   if (r != MAILIMAP_NO_ERROR)
     return r;
   
@@ -202,7 +232,7 @@ static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer,
   if (r != MAILIMAP_NO_ERROR)
     return r;
   
-  r = mailimap_append_uid_parse(fd, buffer, &cur_token, &set);
+  r = mailimap_append_uid_parse(fd, buffer, parser_ctx, &cur_token, &set);
   if (r != MAILIMAP_NO_ERROR)
     return r;
 
@@ -212,7 +242,7 @@ static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer,
     return MAILIMAP_ERROR_MEMORY;
   }
   
-  * index = cur_token;
+  * indx = cur_token;
   * result = resp_code_apnd;
   
   return MAILIMAP_NO_ERROR;
@@ -222,8 +252,8 @@ static int mailimap_resp_code_apnd_parse(mailstream * fd, MMAPString * buffer,
    resp-code-copy  = "COPYUID" SP nz-number SP uid-set SP uid-set
 */
 
-static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index,
+static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx,
     struct mailimap_uidplus_resp_code_copy ** result)
 {
   int r;
@@ -234,7 +264,7 @@ static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
   struct mailimap_uidplus_resp_code_copy * resp_code_copy;
   int res;
   
-  cur_token = * index;
+  cur_token = * indx;
 
   r = mailimap_token_case_insensitive_parse(fd, buffer, &cur_token,
       "COPYUID");
@@ -249,7 +279,7 @@ static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
     goto err;
   }
   
-  r = mailimap_nz_number_parse(fd, buffer, &cur_token, &uidvalidity);
+  r = mailimap_nz_number_parse(fd, buffer, parser_ctx, &cur_token, &uidvalidity);
   if (r != MAILIMAP_NO_ERROR) {
     res = r;
     goto err;
@@ -261,7 +291,7 @@ static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
     goto err;
   }
   
-  r = mailimap_uid_set_parse(fd, buffer, &cur_token, &source_set);
+  r = mailimap_uid_set_parse(fd, buffer, parser_ctx, &cur_token, &source_set);
   if (r != MAILIMAP_NO_ERROR) {
     res = r;
     goto err;
@@ -273,7 +303,7 @@ static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
     goto free_source_set;
   }
   
-  r = mailimap_uid_set_parse(fd, buffer, &cur_token, &dest_set);
+  r = mailimap_uid_set_parse(fd, buffer, parser_ctx, &cur_token, &dest_set);
   if (r != MAILIMAP_NO_ERROR) {
     res = r;
     goto free_source_set;
@@ -286,7 +316,7 @@ static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
     goto free_dest_set;
   }
   
-  * index = cur_token;
+  * indx = cur_token;
   * result = resp_code_copy;
   
   return MAILIMAP_NO_ERROR;
@@ -303,10 +333,10 @@ static int mailimap_resp_code_copy_parse(mailstream * fd, MMAPString * buffer,
   "UIDNOTSTICKY"
 */
 
-static int mailimap_uidplus_uidnotsticky_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index)
+static int mailimap_uidplus_uidnotsticky_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx)
 {
-  return mailimap_token_case_insensitive_parse(fd, buffer, index,
+  return mailimap_token_case_insensitive_parse(fd, buffer, indx,
       "UIDNOTSTICKY");
 }
 
@@ -317,8 +347,8 @@ static int mailimap_uidplus_uidnotsticky_parse(mailstream * fd, MMAPString * buf
                      ; that appears in [IMAP]
 */
 
-static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer,
-    size_t * index,
+static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer, struct mailimap_parser_context * parser_ctx,
+    size_t * indx,
     struct mailimap_extension_data ** result)
 {
   struct mailimap_uidplus_resp_code_apnd * resp_code_apnd;
@@ -327,11 +357,11 @@ static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer
   struct mailimap_extension_data * ext;
   int r;
   
-  cur_token = * index;
+  cur_token = * indx;
   
   resp_code_apnd = NULL;
   resp_code_copy = NULL;
-  r = mailimap_resp_code_apnd_parse(fd, buffer, &cur_token, &resp_code_apnd);
+  r = mailimap_resp_code_apnd_parse(fd, buffer, parser_ctx, &cur_token, &resp_code_apnd);
   if (r == MAILIMAP_NO_ERROR) {
     ext = mailimap_extension_data_new(&mailimap_extension_uidplus,
         MAILIMAP_UIDPLUS_RESP_CODE_APND, resp_code_apnd);
@@ -340,14 +370,14 @@ static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer
       return MAILIMAP_ERROR_MEMORY;
     }
     
-    * index = cur_token;
+    * indx = cur_token;
     * result = ext;
     
     return MAILIMAP_NO_ERROR;
   }
   
   resp_code_copy = NULL;
-  r = mailimap_resp_code_copy_parse(fd, buffer, &cur_token, &resp_code_copy);
+  r = mailimap_resp_code_copy_parse(fd, buffer, parser_ctx, &cur_token, &resp_code_copy);
   if (r == MAILIMAP_NO_ERROR) {
     ext = mailimap_extension_data_new(&mailimap_extension_uidplus,
         MAILIMAP_UIDPLUS_RESP_CODE_COPY, resp_code_copy);
@@ -356,13 +386,13 @@ static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer
       return MAILIMAP_ERROR_MEMORY;
     }
     
-    * index = cur_token;
+    * indx = cur_token;
     * result = ext;
     
     return MAILIMAP_NO_ERROR;
   }
   
-  r = mailimap_uidplus_uidnotsticky_parse(fd, buffer, &cur_token);
+  r = mailimap_uidplus_uidnotsticky_parse(fd, buffer, parser_ctx, &cur_token);
   if (r == MAILIMAP_NO_ERROR) {
     ext = mailimap_extension_data_new(&mailimap_extension_uidplus,
         MAILIMAP_UIDPLUS_RESP_CODE_UIDNOTSTICKY, resp_code_copy);
@@ -371,7 +401,7 @@ static int mailimap_uidplus_resp_code_parse(mailstream * fd, MMAPString * buffer
       return MAILIMAP_ERROR_MEMORY;
     }
     
-    * index = cur_token;
+    * indx = cur_token;
     * result = ext;
     
     return MAILIMAP_NO_ERROR;
