@@ -40,6 +40,7 @@
 #endif
 
 #include "mailsmtp.h"
+#include "mailsmtp_private.h"
 #include "connect.h"
 #include "base64.h"
 #include "mail.h"
@@ -98,6 +99,9 @@
      Messages
 */
 
+static inline void smtp_logger(mailstream * s, int log_type,
+    const char * str, size_t size, void * context);
+
 #define SMTP_STATUS_CONTINUE 0x1000
 
 mailsmtp * mailsmtp_new(size_t progr_rate,
@@ -128,9 +132,7 @@ mailsmtp * mailsmtp_new(size_t progr_rate,
   session->esmtp = 0;
   session->auth = MAILSMTP_AUTH_NOT_CHECKED;
   
-#ifdef USE_SASL
   session->smtp_sasl.sasl_conn = NULL;
-#endif
   
   session->smtp_max_msg_size = 0;
   session->smtp_progress_fun = NULL;
@@ -138,6 +140,9 @@ mailsmtp * mailsmtp_new(size_t progr_rate,
 
 	session->smtp_timeout = 0;
 
+  session->smtp_logger = NULL;
+  session->smtp_logger_context = NULL;
+  
   return session;
 
  free_line_buffer:
@@ -179,6 +184,7 @@ int mailsmtp_connect(mailsmtp * session, mailstream * s)
   int code;
 
   session->stream = s;
+  mailstream_set_logger(s, smtp_logger, session);
 
   code = read_response(session);
 
@@ -190,7 +196,12 @@ int mailsmtp_connect(mailsmtp * session, mailstream * s)
     session->stream = NULL;
     mailstream_close(s);
     return MAILSMTP_ERROR_SERVICE_NOT_AVAILABLE;
-      
+
+  case 0:
+    session->stream = NULL;
+    mailstream_close(s);
+    return MAILSMTP_ERROR_STREAM;
+
   default:
     session->stream = NULL;
     mailstream_close(s);
@@ -265,7 +276,7 @@ static int get_hostname(mailsmtp * session, int useip, char * buf, int len)
     if (r != 0)
       return MAILSMTP_ERROR_HOSTNAME;
 
-#if (defined __linux__ || defined WIN32)
+#if (defined __linux__ || defined WIN32 || defined __sun)
     r = getnameinfo(&addr, sizeof(addr), hostname, HOSTNAME_SIZE, NULL, 0, NI_NUMERICHOST);
 #else
     r = getnameinfo(&addr, addr.sa_len, hostname, HOSTNAME_SIZE, NULL, 0, NI_NUMERICHOST);
@@ -391,6 +402,9 @@ int mailsmtp_data(mailsmtp * session)
   case 503:
     return MAILSMTP_ERROR_BAD_SEQUENCE_OF_COMMAND;
 
+  case 0:
+    return MAILSMTP_ERROR_STREAM;
+
   default:
     return MAILSMTP_ERROR_UNEXPECTED_CODE;
   }
@@ -434,45 +448,112 @@ int mailsmtp_data_message(mailsmtp * session,
   }
 }
 
+int mailsmtp_status(int smtpstatus)
+{
+  switch(smtpstatus) {
+  case 250:
+    return MAILSMTP_NO_ERROR;
+
+  case 552:
+    return MAILSMTP_ERROR_EXCEED_STORAGE_ALLOCATION;
+
+  case 554:
+    return MAILSMTP_ERROR_TRANSACTION_FAILED;
+
+  case 451:
+    return MAILSMTP_ERROR_IN_PROCESSING;
+
+  case 452:
+    return MAILSMTP_ERROR_INSUFFICIENT_SYSTEM_STORAGE;
+
+  case 0:
+    return MAILSMTP_ERROR_STREAM;
+
+  default:
+    return MAILSMTP_ERROR_UNEXPECTED_CODE;
+  }
+
+}
+
+/* lmtp data command
+ * on the end of data transaction lmtp returns a status for
+ * every recipient it delivered.
+ * recipient_list holds the rcpts from smtp_rcpt()
+ * the last error is returned to the caller
+ * if retcodes is not NULL, the individual rcpt-stati are stored there
+ *  */
+int maillmtp_data_message(mailsmtp * session,
+                          const char * message,
+                          size_t size,
+                          clist * recipient_list,
+                          int * retcodes)
+{
+  int r, ret = MAILSMTP_NO_ERROR;
+  unsigned int i = 0;
+  clistiter *iter;
+
+  r = send_data(session, message, size);
+
+  if (r == -1)
+    return MAILSMTP_ERROR_STREAM;
+
+  for (iter = clist_begin(recipient_list); iter; iter = clist_next(iter)) {
+    r = read_response(session);
+    if (MAILSMTP_NO_ERROR != mailsmtp_status(r))
+      ret = mailsmtp_status(r);
+    if (retcodes)
+      retcodes[i++] = r;
+  }
+  return ret;
+}
+
 int mailsmtp_data_message_quit(mailsmtp * session,
                                const char * message,
                                size_t size)
 {
-    int r;
+  int r = mailsmtp_data_message_quit_no_disconnect(session, message, size);
+  
+  mailstream_close(session->stream);
+  session->stream = NULL;
     
-    r = send_data(session, message, size);
-    if (r == -1)
-        return MAILSMTP_ERROR_STREAM;
-    
-    r = send_quit(session);
-    
-    r = read_response(session);
-    
-    mailstream_close(session->stream);
-    session->stream = NULL;
-    
-    switch(r) {
-        case 250:
-            return MAILSMTP_NO_ERROR;
-            
-        case 552:
-            return MAILSMTP_ERROR_EXCEED_STORAGE_ALLOCATION;
-            
-        case 554:
-            return MAILSMTP_ERROR_TRANSACTION_FAILED;
-            
-        case 451:
-            return MAILSMTP_ERROR_IN_PROCESSING;
-            
-        case 452:
-            return MAILSMTP_ERROR_INSUFFICIENT_SYSTEM_STORAGE;
-            
-        case 0:
-            return MAILSMTP_ERROR_STREAM;
-            
-        default:
-            return MAILSMTP_ERROR_UNEXPECTED_CODE;
-    }
+  return r;
+}
+
+int mailsmtp_data_message_quit_no_disconnect(mailsmtp * session,
+                                             const char * message,
+                                             size_t size)
+{
+  int r;
+  
+  r = send_data(session, message, size);
+  if (r == -1)
+    return MAILSMTP_ERROR_STREAM;
+  
+  r = send_quit(session);
+  r = read_response(session);
+  
+  switch(r) {
+    case 250:
+      return MAILSMTP_NO_ERROR;
+      
+    case 552:
+      return MAILSMTP_ERROR_EXCEED_STORAGE_ALLOCATION;
+      
+    case 554:
+      return MAILSMTP_ERROR_TRANSACTION_FAILED;
+      
+    case 451:
+      return MAILSMTP_ERROR_IN_PROCESSING;
+      
+    case 452:
+      return MAILSMTP_ERROR_INSUFFICIENT_SYSTEM_STORAGE;
+      
+    case 0:
+      return MAILSMTP_ERROR_STREAM;
+      
+    default:
+      return MAILSMTP_ERROR_UNEXPECTED_CODE;
+  }
 }
 
 /* esmtp operations */
@@ -529,6 +610,9 @@ int mailesmtp_parse_ehlo(mailsmtp * session)
       }
       /* TODO: grab optionnal max size */
     }
+    else if (!strncasecmp(response, "CLIENTID", 8) && isdelim(response[8])) {
+      session->esmtp |= MAILSMTP_ESMTP_CLIENTID;
+    }
     else if (!strncasecmp(response, "PIPELINING", 10) && isdelim(response[10])) {
       session->esmtp |= MAILSMTP_ESMTP_PIPELINING;
     }
@@ -651,6 +735,39 @@ int mailesmtp_ehlo_with_ip(mailsmtp * session, int useip)
   }
 }
 
+/* lmtp processing */
+int mailesmtp_lhlo(mailsmtp * session , const char *hostname)
+{
+  int r;
+  char command[SMTP_STRING_SIZE];
+
+  if (!hostname)
+      hostname = "localhost";
+
+  snprintf(command, SMTP_STRING_SIZE, "LHLO %s\r\n", hostname);
+  r = send_command(session, command);
+  if (r == -1)
+    return MAILSMTP_ERROR_STREAM;
+  r = read_response(session);
+
+  switch (r) {
+  case 250:
+    return mailesmtp_parse_ehlo(session);
+
+  case 504:
+    return MAILSMTP_ERROR_NOT_IMPLEMENTED;
+
+  case 550:
+    return MAILSMTP_ERROR_ACTION_NOT_TAKEN;
+
+  case 0:
+    return MAILSMTP_ERROR_STREAM;
+
+  default:
+    return MAILSMTP_ERROR_UNEXPECTED_CODE;
+  }
+}
+
 /*
   if return_full is TRUE, the entire message is returned on error
   envid can be NULL
@@ -951,6 +1068,32 @@ int mailsmtp_auth(mailsmtp * session, const char * user, const char * pass)
   }
 }
 
+int mailesmtp_clientid(mailsmtp * session,
+    const char * type, const char * clientid) {
+  char command[SMTP_STRING_SIZE];
+  int r;
+
+  if (!(session->esmtp & MAILSMTP_ESMTP_CLIENTID))
+    return MAILSMTP_ERROR_CLIENTID_NOT_SUPPORTED;
+
+  snprintf(command, SMTP_STRING_SIZE, "CLIENTID %s %s\r\n",
+          type, clientid);
+
+  r = send_command(session, command);
+  if (r == -1)
+    return MAILSMTP_ERROR_STREAM;
+  r = read_response(session);
+
+  switch (r) {
+  case 250:
+    return MAILSMTP_NO_ERROR;
+  case 501: // syntax error
+  case 503: // duplicate clientid or not permitted because of policy
+  default:
+    return MAILSMTP_ERROR_UNEXPECTED_CODE;
+  }
+}
+
 /* TODO: add mailesmtp_etrn, mailssmtp_expn */
 
 int mailesmtp_starttls(mailsmtp * session)
@@ -972,6 +1115,9 @@ int mailesmtp_starttls(mailsmtp * session)
   case 454:
     return MAILSMTP_ERROR_STARTTLS_TEMPORARY_FAILURE;
 
+  case 0:
+    return MAILSMTP_ERROR_STREAM;
+
   default:
     return MAILSMTP_ERROR_UNEXPECTED_CODE;
   }
@@ -984,7 +1130,7 @@ static int parse_response(mailsmtp * session,
   int code;
   int cont = 0;
 
-  code = strtol(response, &message, 10);
+  code = (int) strtol(response, &message, 10);
   if (* message == ' ')
     mmap_string_append(session->response_buffer, message + 1);
   else if (* message == '-') {
@@ -1107,6 +1253,8 @@ const char * mailsmtp_strerror(int errnum)
     return "Transaction failed";
   case MAILSMTP_ERROR_MEMORY:
     return "Memory error";
+  case MAILSMTP_ERROR_AUTH_NOT_SUPPORTED:
+    return "Authentication is not supported";
   case MAILSMTP_ERROR_CONNECTION_REFUSED:
     return "Connection refused";
   case MAILSMTP_ERROR_STARTTLS_TEMPORARY_FAILURE:
@@ -1114,7 +1262,9 @@ const char * mailsmtp_strerror(int errnum)
   case MAILSMTP_ERROR_STARTTLS_NOT_SUPPORTED:
     return "TLS not supported by server";
   case MAILSMTP_ERROR_AUTH_LOGIN:
-  	return "Login failed";
+    return "Login failed";
+  case MAILSMTP_ERROR_CLIENTID_NOT_SUPPORTED:
+    return "ClientID not supported by server";
   default:
     return "Unknown error code";
   }
@@ -1134,14 +1284,14 @@ static int sasl_getsimple(void * context, int id,
     if (result != NULL)
       * result = session->smtp_sasl.sasl_login;
     if (len != NULL)
-      * len = strlen(session->smtp_sasl.sasl_login);
+      * len = (unsigned) strlen(session->smtp_sasl.sasl_login);
     return SASL_OK;
     
   case SASL_CB_AUTHNAME:
     if (result != NULL)
       * result = session->smtp_sasl.sasl_auth_name;
     if (len != NULL)
-      * len = strlen(session->smtp_sasl.sasl_auth_name);
+      * len = (unsigned) strlen(session->smtp_sasl.sasl_auth_name);
     return SASL_OK;
   }
   
@@ -1294,6 +1444,10 @@ int mailesmtp_auth_sasl(mailsmtp * session, const char * auth_type,
   while (1) {
     r = read_response(session);
     switch (r) {
+    case 0:
+      res = MAILSMTP_ERROR_STREAM;
+      goto free_sasl_conn;
+        
     case 220:
     case 235:
       res = MAILSMTP_NO_ERROR;
@@ -1310,7 +1464,7 @@ int mailesmtp_auth_sasl(mailsmtp * session, const char * auth_type,
     
     case 334:
       {
-        size_t response_len;
+        unsigned int response_len;
         char * decoded;
         unsigned int decoded_len;
         unsigned int max_decoded;
@@ -1325,7 +1479,7 @@ int mailesmtp_auth_sasl(mailsmtp * session, const char * auth_type,
           * p = '\0';
         }
         
-        response_len = strlen(session->response);
+        response_len = (unsigned int) strlen(session->response);
         max_decoded = response_len * 3 / 4;
         decoded = malloc(max_decoded + 1);
         if (decoded == NULL) {
@@ -1452,3 +1606,37 @@ time_t mailsmtp_get_timeout(mailsmtp * session)
 	return session->smtp_timeout;
 }
 
+static inline void smtp_logger(mailstream * s, int log_type,
+    const char * str, size_t size, void * context)
+{
+  mailsmtp * session;
+
+  session = context;
+  if (session->smtp_logger == NULL)
+    return;
+
+  session->smtp_logger(session, log_type, str, size, session->smtp_logger_context);
+}
+
+LIBETPAN_EXPORT
+void mailsmtp_set_logger(mailsmtp * session, void (* logger)(mailsmtp * session, int log_type,
+    const char * str, size_t size, void * context), void * logger_context)
+{
+  session->smtp_logger = logger;
+  session->smtp_logger_context = logger_context;
+}
+
+int mailsmtp_send_command(mailsmtp * f, char * command)
+{
+  return send_command(f, command);
+}
+
+int mailsmtp_send_command_private(mailsmtp * f, char * command)
+{
+    return send_command_private(f, command, 0);
+}
+
+int mailsmtp_read_response(mailsmtp * session)
+{
+  return read_response(session);
+}
